@@ -1,4 +1,5 @@
 """
+
     Copyright (C) 2021 MiranDaniel
 
     This program is free software: you can redistribute it and/or modify
@@ -13,10 +14,13 @@
 
     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
+
 """
 
 import requests
 import json
+from .exceptions import *
+
 
 def _validate_ip(s):
     a = s.split('.')
@@ -31,23 +35,89 @@ def _validate_ip(s):
             return False
     return True
 
+
+def _validate_address(s):
+    if s.startswith("nano_") == False and s.startswith("ban_") == False and s.startswith("xrb_") == False:
+        raise AddressInvalid(s)
+    else:
+        if s.startswith("nano_"):
+            if len(s) != 65:
+                raise AddressInvalid(s)
+        if s.startswith("xrb_"):
+            if len(s) != 65:
+                raise AddressInvalid(s)
+        elif s.startswith("ban_"):
+            if len(s) != 64:
+                raise AddressInvalid(s)
+
+
+def _validate_block(s):
+    if len(s) != 64:
+        raise BlockInvalid(s)
+
+
 class Node:
-    def __init__(self, ip, port=7076, dontUseHTTPS=False, headers="Default"):
-        self.ip = ip
-        self.port = port
-        self.secure = "s" if dontUseHTTPS == False else ""
-        if _validate_ip(self.ip) == True:
-            self.target = f"http{self.secure}://{self.ip}:{self.port}"
+    def __init__(self, ip, port="Default", dontUseHTTPS=False, headers="Default", banano=False):
+        self.ip = ip.lower().replace("https", "").replace("http", "")
+        self.secure = "https" if dontUseHTTPS == False else "http"
+
+        if port == "Default":
+            if banano == False:
+                self.port = 7076
+            else:
+                self.port = 7072
         else:
-            self.target = f"http{self.secure}://{self.ip}"
-        if headers != "Default":
-            self.headers = {'Content-type': 'application/json', 'Accept': '*/*',"Accept-Encoding":"gzip, deflate, br","Connection":"keep-alive"}
+            self.port = port
+
+        self.banano = banano
+
+        if _validate_ip(self.ip) == True:
+            self.target = f"{self.secure}://{self.ip}:{self.port}"
+        else:
+            self.target = f"{self.secure}://{self.ip}"
+
+        if headers == "Default":
+            self.headers = {'Content-type': 'application/json', 'Accept': '*/*',
+                            "Accept-Encoding": "gzip, deflate, br", "Connection": "keep-alive"}
+            if banano:
+                self.headers = {
+                    'Content-type': 'application/json', "Connection": "keep-alive"}
         else:
             self.headers = headers
 
     def _request(self, data):
-        response = requests.post(self.target, data=json.dumps(data), headers=self.headers).text
-        return json.loads(response)
+        if "account" in data:
+            _validate_address(data["account"])
+        if "block" in data:
+            _validate_block(data["block"])
+        try:
+            response = requests.post(self.target, data=json.dumps(
+                data), headers=self.headers).text
+        except requests.exceptions.ConnectionError:
+            raise CannotConnect()
+        if response.startswith("<!DOCTYPE html>"):
+            raise InvalidServerResponseHTML()
+        jsload = json.loads(response)
+        if "message" in jsload:
+            if jsload["message"] == "Action is not supported":
+                raise ActionNotSupported()
+        if "error" in jsload:
+            if jsload["error"] == "Wallet not found":
+                raise WalletNotFound()
+            elif jsload["error"] == "Invalid block hash":
+                raise InvalidBlockHash()
+            elif jsload["error"] == "Unknown error":
+                raise UnknownError()
+            elif jsload["error"] == "RPC control is disabled":
+                raise RPCdisabled()
+            elif jsload["error"] == "Invalid balance number":
+                raise InvalidBalanceNumber()
+            elif jsload["error"] == "Empty response":
+                raise EmptyResponse()
+            else:
+                raise UnknownError()
+
+        return jsload
 
     def account_balance(self, account):
         response = self._request(
@@ -206,8 +276,15 @@ class Node:
             "subtype": response["subtype"],
         }
 
-    # blocks
-    # blocks_info
+    def blocks_info(self, json_block, hashes):
+        response = self._request(
+            {"action": "blocks_info", "json_block": json_block, "hashes": hashes})
+        return response["blocks"]
+
+    def blocks(self, json_block, hashes):
+        response = self._request(
+            {"action": "blocks", "json_block": json_block, "hashes": hashes})
+        return response["blocks"]
 
     def bootstrap(self, address, port):
         response = self._request(
@@ -241,7 +318,7 @@ class Node:
             "attempts": []
         }
         for i in response["attempts"]:
-            ret["attempts"].append(i)  # WORK ON DATATYPE CONVERSION
+            ret["attempts"].append(i)
 
         return ret
 
@@ -310,7 +387,7 @@ class Node:
 
     def deterministic_key(self, seed, index):
         response = self._request(
-            {"action": "deterministic_key", "seed":seed,"index":index})
+            {"action": "deterministic_key", "seed": seed, "index": index})
         return response
 
     def epoch_upgrade(self, epoch, key):
@@ -322,8 +399,9 @@ class Node:
         response = self._request({"action": "frontier_count"})
         return int(response["count"])
 
-    def frontiers(self,account,count):
-        response = self._request({"action": "frontiers","account":account,"count":count})
+    def frontiers(self, account, count):
+        response = self._request(
+            {"action": "frontiers", "account": account, "count": count})
         return response["frontiers"]
 
     def keepalive(self, address, port):
@@ -340,7 +418,8 @@ class Node:
         return response
 
     def ledger(self, account, count):
-        response = self._request({"action": "ledger", "account":account,"count":count})
+        response = self._request(
+            {"action": "ledger", "account": account, "count": count})
         ret = {}
         for i in response["accounts"]:
             ret[i] = {
@@ -378,7 +457,8 @@ class Node:
         return int(response["exists"])
 
     def process(self, json_block, subtype, block):
-        response = self._request({"action": "process", "json_block":json_block,"subtype":subtype, "block":block})
+        response = self._request(
+            {"action": "process", "json_block": json_block, "subtype": subtype, "block": block})
         return response["hash"]
 
     def representatives(self):
@@ -458,15 +538,58 @@ class Node:
             "build_info": response["build_info"]
         }
 
-    # unchecked
+    def unchecked(self, json_block, count):
+        response = self._request(
+            {"action": "unchecked", "json_block": json_block, "count": count})
+        ret = {}
+        for i in response["blocks"]:
+            ret[i] = {
+                "type": response["blocks"][i]["type"],
+                "account": response["blocks"][i]["account"],
+                "previous": response["blocks"][i]["previous"],
+                "representative": response["blocks"][i]["representative"],
+                "balance": int(response["blocks"][i]["balance"]),
+                "link": response["blocks"][i]["link"],
+                "link_as_account": response["blocks"][i]["link_as_account"],
+                "signature": response["blocks"][i]["signature"],
+                "work": response["blocks"][i]["work"]
+            }
+        return ret
 
     def unchecked_clear(self):
         response = self._request({"action": "unchecked_clear"})
         return response["success"]
 
-    # unchecked_get
-    # unchecked_keys
-    # unopened
+    def unchecked_get(self, json_block, Hash):
+        response = self._request(
+            {"action": "unchecked_get", "json_block": json_block, "hash": Hash})
+        return {
+            "modified_timestamp": int(response["modified_timestamp"]),
+            "contents": {
+                "type": response["contents"]["type"],
+                "account": response["contents"]["account"],
+                "previous": response["contents"]["previous"],
+                "representative": response["contents"]["representative"],
+                "balance": int(response["contents"]["balance"]),
+                "link": response["contents"]["link"],
+                "link_as_account": response["contents"]["link_as_account"],
+                "signature": response["contents"]["signature"],
+                "work": response["contents"]["work"]
+            }
+        }
+
+    def unchecked_keys(self, json_block, key, count):
+        response = self._request(
+            {"action": "unchecked_keys", "json_block": json_block, "key": key, "count": count})
+        return response["unchecked"]
+
+    def unopened(self, account, count):
+        response = self._request(
+            {"action": "unopened", "account": account, "count": count})
+        ret = {}
+        for i in response["accounts"]:
+            ret[i] = int(response["accounts"][i])
+        return ret
 
     def uptime(self):
         response = self._request({"action": "uptime"})
@@ -476,7 +599,14 @@ class Node:
         response = self._request({"action": "work_cancel", "hash": Hash})
         return response
 
-    # work_generate
+    def work_generate(self, Hash):
+        response = self._request({"action": "work_generate", "hash": Hash})
+        return {
+            "work": response["work"],
+            "difficulty": response["difficulty"],
+            "multiplier": float(response["multiplier"]),
+            "hash": response["hash"]
+        }
 
     def work_peer_add(self, address, port):
         response = self._request(
@@ -491,7 +621,15 @@ class Node:
         response = self._request({"action": "work_peer_add"})
         return response["success"]
 
-    # work validate
+    def work_validate(self, work, Hash):
+        response = self._request(
+            {"action": "work_validate", "work": work, "hash": Hash})
+        return {
+            "valid_all": int(response["valid_all"]),
+            "valid_receive": int(response["valid_receive"]),
+            "difficulty": response["difficulty"],
+            "multiplier": float(response["multiplier"])
+        }
 
     def account_create(self, wallet):
         response = self._request(
@@ -614,7 +752,10 @@ class Node:
             {"action": "wallet_destroy", "wallet": wallet})
         return int(response["destroyed"])
 
-    # wallet_export
+    def wallet_export(self, wallet):
+        response = self._request(
+            {"action": "wallet_export", "wallet": wallet})
+        return json.loads(response["json"])
 
     def wallet_frontiers(self, wallet):
         response = self._request(
